import re
import datetime
from urllib.parse import unquote
from cgi_parsing import print_svg_response
from cytoband_utils import retrieve_gene_id_coordinates

# http://progenetix.org/cgi/bycon/services/intervalFrequencies.py?chr2plot=8,9,17&labels=8:120000000-123000000:Some+Interesting+Region&plot_gene_symbols=MYCN,REL,TP53,MTAP,CDKN2A,MYC,ERBB2,CDK1&filters=pgx:icdom-85003&output=histoplot

################################################################################

def histoplot_svg_generator(byc, results):

    if not "histoplot" in byc["output"]:
        return byc

    plv = _initialize_plot_values(byc, results)

    _plot_add_title(plv, byc)
    _plot_add_cytobands(plv, byc)
    _plot_add_histodata(plv, results, byc)
    _plot_add_markers(plv, byc)
    _plot_add_footer(plv, results, byc)

    plv["Y"] += plv["plot_margins"]

    #--------------------------------------------------------------------------#

    svg = _create_svg(plv)
    print_svg_response(svg, byc["env"])

################################################################################

def _create_svg(plv):

    svg = """<svg
xmlns="http://www.w3.org/2000/svg"
xmlns:xlink="http://www.w3.org/1999/xlink"
version="1.1"
id="{}"
width="{}px"
height="{}px"
style="margin: auto; font-family: Helvetica, sans-serif;">
<style type="text/css"><![CDATA[
{}
]]></style>
<rect x="0" y="0" width="{}" height="{}" style="fill: {}" />
{}
</svg>""".format(
        plv["plot_id"],
        plv["plot_width"],
        plv["Y"],
        "\n  ".join(plv["styles"]),
        plv["plot_width"],
        plv["Y"],
        plv["plot_canvas_color"],
        "\n".join(plv["pls"])
    )

    return svg

################################################################################

def _initialize_plot_values(byc, results):

    p_d_p = byc["plot_defaults"]["parameters"]
    r_no = len(results)

    plv = {
        "pls": [],
        "histogram_number": r_no
    }

    for p_k, p_d in p_d_p.items():
        plv.update({ p_k: p_d.get("default", "") })

    if plv["histogram_number"] < 2:
        plv.update({ "plot_labelcol_width": 0 })

    _plot_get_form_parameters(plv, byc)

    pax = plv["plot_margins"] + plv["plot_labelcol_width"] + plv["plot_axislab_y_width"]

    paw = plv["plot_width"] - 2 * plv["plot_margins"]
    paw -= plv["plot_labelcol_width"]
    paw -= plv["plot_axislab_y_width"]
    paw -= plv["plot_treecol_width"]

    # calculate the base
    chr_b_s = 0
    for chro in plv["plot_chros"]:
        c_l = byc["cytolimits"][chro]
        chr_b_s += c_l["size"]

    pyf = plv["plot_area_height"] * 0.5 / plv["plot_axis_y_max"]

    gaps = len(plv["plot_chros"]) - 1
    gap_sw = gaps * plv["plot_region_gap_width"]
    genome_width = paw - gap_sw
    b2pf = genome_width / chr_b_s # TODO: only exists if using stack

    title = plv.get("plot_title", "")
    if len(title) < 3:
        if plv["histogram_number"] == 1:
            title = _format_resultset_title(results[0])

    plv.update({
        "plot_title": title,
        "styles": [],
        "Y": plv["plot_margins"],
        "plot_area_width": paw,
        "plot_area_x0": pax,
        "plot_area_xe": pax + paw,
        "plot_area_xc": pax + paw / 2,
        "plot_y2pf": pyf,
        "plot_genome_size": chr_b_s,
        "plot_b2pf": b2pf,
        "plot_labels": {}
    })

    return plv

################################################################################

def _format_resultset_title(f_set):

    g_id = f_set.get("group_id", "NA")
    g_lab = f_set.get("label", "")
    if len(g_lab) > 1:
        title = "{} ({})".format(g_lab, g_id)
    else:
        title = g_id

    return title

################################################################################

def _plot_get_form_parameters(plv, byc):

    p_d_p = byc["plot_defaults"]["parameters"]
    p_d_l = byc["plot_defaults"]["legacy_parameters"]
    form = byc["form_data"]

    for p_k, p_d in p_d_p.items():
        if p_k in form:
            plv.update({ p_k: form[p_k]})
        l_p = p_d_l.get(p_k, False)
        if l_p is not False:
            if l_p in byc["form_data"]:
                plv.update({ p_k: form[l_p]})

        if "int" in p_d.get("type", "string"):
            plv.update({ p_k: int(plv[p_k])})
        
        elif "array" in p_d.get("type", "string"):
            if isinstance(plv[p_k], str):
                plv.update({ p_k: unquote(plv[p_k]) })
                plv.update({ p_k: re.split(",", plv[p_k])})
            if "int" in p_d.get("items", "string"):
                plv.update({ p_k: list(map(int, plv[p_k])) })

    return plv

################################################################################

def _plot_add_title(plv, byc):

    if len(plv.get("plot_title", "")) < 3:
        return plv

    t_l = len(plv["plot_title"]) * 0.45 * plv["plot_title_font_size"]
    plv["Y"] += plv["plot_title_font_size"]

    plv["pls"].append(
'<text x="{}" y="{}" style="text-anchor: middle; font-size: {}px">{}</text>'.format(
            plv["plot_area_xc"],
            plv["Y"],
            plv["plot_title_font_size"],
            plv["plot_title"]
        )
    )

    plv["Y"] += plv["plot_title_font_size"]

    return plv

################################################################################

def _plot_add_cytobands(plv, byc):

    if plv["plot_chro_height"] < 1:
        return plv

    _plot_add_cytoband_svg_gradients(plv)

    #------------------------- chromosome labels ------------------------------#

    X = plv["plot_area_x0"]
    plv["Y"] += plv["plot_title_font_size"]

    for chro in plv["plot_chros"]:

        c_l = byc["cytolimits"][str(chro)]

        chr_w = c_l["size"] * plv["plot_b2pf"]
        chr_c = X + chr_w / 2

        plv["pls"].append(f'<text x="{chr_c}" y="{plv["Y"]}" style="text-anchor: middle; font-size: {plv["plot_font_size"]}px">{chro}</text>')

        X += chr_w
        X += plv["plot_region_gap_width"]

    plv["Y"] += plv["plot_region_gap_width"]

    #---------------------------- chromosomes ---------------------------------#

    X = plv["plot_area_x0"]

    for chro in plv["plot_chros"]:

        c_l = byc["cytolimits"][str(chro)]
        chr_w = c_l["size"] * plv["plot_b2pf"]

        chr_cb_s = list(filter(lambda d: d[ "chro" ] == chro, byc["cytobands"].copy()))

        last = len(chr_cb_s) - 1
        this_n = 0

        for cb in chr_cb_s:

            this_n += 1

            s_b = cb["start"]
            e_b = cb["end"]
            c = cb["staining"]
            l = int(e_b) - int(s_b)
            l_px = l * plv["plot_b2pf"]

            by = plv["Y"]
            bh = plv["plot_chro_height"]

            if "cen" in c:
                by += 0.2 * plv["plot_chro_height"]
                bh -= 0.4 * plv["plot_chro_height"]
            elif "stalk" in c:
                by += 0.3 * plv["plot_chro_height"]
                bh -= 0.6 * plv["plot_chro_height"]
            elif this_n == 1 or this_n == last:
                by += 0.1 * plv["plot_chro_height"]
                bh -= 0.2 * plv["plot_chro_height"]

            plv["pls"].append(f'<rect x="{round(X, 1)}" y="{round(by, 1)}" width="{round(l_px, 1)}" height="{round(bh, 1)}" style="fill: url(#{plv["plot_id"]}{c}); " />')

            X += l_px

        X += plv["plot_region_gap_width"]

    #-------------------------- / chromosomes ---------------------------------#

    plv["Y"] += plv["plot_chro_height"]
    plv["Y"] += plv["plot_region_gap_width"]

    return plv

################################################################################

def _plot_add_histodata(plv, results, byc):

    plv.update( {"plot_first_histo_y0": plv["Y"] })

    for f_set in results:
        _plot_add_one_histogram(plv, f_set, byc)

    return plv

################################################################################

def _plot_add_one_histogram(plv, f_set, byc):

    _plot_add_one_histogram_canvas(plv, f_set, byc)

    i_f = f_set.get("interval_frequencies", [])

    X = plv["plot_area_x0"]
    h_y_0 = plv["Y"] + plv["plot_area_height"] * 0.5

    #------------------------- histogram data ---------------------------------#

    # TODO: in contrast to the Perl version here we don't correct for interval
    #       sets which _do not_ correspond to the full chromosome coordinates

    for chro in plv["plot_chros"]:

        c_l = byc["cytolimits"][str(chro)]
        chr_w = c_l["size"] * plv["plot_b2pf"]

        c_i_f = list(filter(lambda d: d[ "reference_name" ] == chro, i_f.copy()))
        c_i_no = len(c_i_f)

        for GL in ["gain_frequency", "loss_frequency"]:

            if "gain_frequency" in GL:
                h_f = -1
                p_c = plv["plot_dup_color"]
            else:
                h_f = 1
                p_c = plv["plot_del_color"]

            p = f'<polygon points="\n{ round(X, 1) } { round(h_y_0, 1) }'

            i_x_0 = X
            prev = -1
            for c_i_i, i_v in enumerate(c_i_f, start=1):

                s = i_x_0 + i_v.get("start", 0) * plv["plot_b2pf"]
                e = i_x_0 + i_v.get("end", 0) * plv["plot_b2pf"]
                v = i_v.get(GL, 0)
                h = v * plv["plot_y2pf"]
                h_p = h_y_0 + h * h_f

                point = f'\n{ round(s, 1) } {round(h_p, 1)}'

                # This construct avoids adding intermediary points w/ the same
                # value as the one before and after 
                if c_i_no > c_i_i:
                    future = c_i_f[c_i_i].get(GL, 0)
                    if prev != v or future != v:
                        p += point
                else:
                    p += point

                prev = v

            p += f'\n{ round((X + chr_w), 1) } { round(h_y_0, 1) }"' 
            p += f'\nfill="{p_c}" stroke-width="0px" />'

            plv["pls"].append(p)

            i_x_0 += i_v.get("start", 0) * plv["plot_b2pf"]

        X += chr_w
        X += plv["plot_region_gap_width"]

    #------------------------ / histogram data --------------------------------#

    plv["Y"] += plv["plot_area_height"]
    plv.update( {"plot_last_histo_ye": plv["Y"] })
    plv["Y"] += plv["plot_region_gap_width"]

    return plv

################################################################################

def _plot_add_one_histogram_canvas(plv, f_set, byc):

    x_a_0 = plv["plot_area_x0"]
    p_a_w = plv["plot_area_width"]
    p_a_h = plv["plot_area_height"]
    p_a_c = plv["plot_area_color"]

    #-------------------------- left labels -----------------------------------#
    _plot_add_left_labels(plv, f_set, byc)

    #--------------------- plot area background -------------------------------#
    
    plv["pls"].append(f'<rect x="{x_a_0}" y="{plv["Y"]}" width="{p_a_w}" height="{p_a_h}" style="fill: {p_a_c}; fill-opacity: 0.8; " />')

    #--------------------------- grid lines -----------------------------------#
    _plot_area_add_grid(plv, byc)

    return plv

################################################################################

def _plot_add_left_labels(plv, f_set, byc):

    if plv["histogram_number"] < 2:
        return plv

    lab_x_e = plv["plot_margins"] + plv["plot_labelcol_width"]
    h_y_0 = plv["Y"] + plv["plot_area_height"] * 0.5

    plv["styles"].append(
        f'.title-left {{text-anchor: end; fill: { plv["plot_font_color"] }; font-size: { plv["plot_font_size"] }px;}}'
    )
    
    g_id = f_set.get("group_id", "NA")
    g_lab = f_set.get("label", "")
    g_no = f_set.get("sample_count", 0)

    # The condition splits the label data on 2 lines if a text label pre-exists
    if len(g_lab) > 0:
        lab_y = h_y_0 - plv["plot_labelcol_font_size"] * 0.2
        plv["pls"].append(f'<text x="{lab_x_e}" y="{lab_y}" class="title-left">{g_lab}</text>')
        lab_y = h_y_0 + plv["plot_labelcol_font_size"] * 1.2
        plv["pls"].append(f'<text x="{lab_x_e}" y="{lab_y}" class="title-left">{g_id} ({g_no} samples)</text>')
    else:
        lab_y = h_y_0 - plv["plot_labelcol_font_size"] * 0.5
        plv["pls"].append(f'<text x="{lab_x_e}" y="{lab_y}" class="title-left">{g_id} ({g_no} samples)</text>')

    return plv

################################################################################

def _plot_area_add_grid(plv, byc):

    x_a_0 = plv["plot_area_x0"]
    x_c_e = plv["plot_area_xe"]

    h_y_0 = plv["Y"] + plv["plot_area_height"] * 0.5
    x_y_l = x_a_0 - plv["plot_region_gap_width"]

    plv["styles"].append(
        f'.label-y {{text-anchor: end; fill: { plv["plot_label_y_font_color"] }; font-size: {plv["plot_label_y_font_size"]}px;}}'
    )   
    plv["styles"].append(
        f'.gridline {{stroke-width: {plv["plot_grid_stroke"]}px; stroke: {plv["plot_grid_color"]}; opacity: {plv["plot_grid_opacity"]} ; }}',
    )   

    #-------------------------- center line -----------------------------------#

    plv["pls"].append(f'<line x1="{x_a_0}"  y1="{h_y_0}"  x2="{x_c_e}"  y2="{h_y_0}" class="gridline" />')

    #--------------------------- grid lines -----------------------------------#

    for y_m in plv["plot_histogram_frequency_labels"]:

        if y_m > plv["plot_axis_y_max"]:
            continue

        for f in [1, -1]:

            y_v = h_y_0 + f * y_m * plv["plot_y2pf"]
            y_l_y = y_v + plv["plot_label_y_font_size"] / 2

            plv["pls"].append(f'<line x1="{x_a_0}" y1="{y_v}" x2="{x_c_e}" y2="{y_v}" class="gridline" />')
            plv["pls"].append(f'<text x="{x_y_l}" y="{y_l_y}" class="label-y">{y_m}%</text>')

    return plv

################################################################################

def _plot_add_markers(plv, byc):

    _add_labs_from_plot_region_labels(plv, byc)
    _add_labs_from_gene_symbols(plv, byc)

    labs = plv["plot_labels"]

    if not labs:
        return plv

    b2pf = plv["plot_b2pf"]

    max_lane = 0
    marker_y_0 = round(plv["plot_first_histo_y0"] - 1, 1)
    marker_y_e = round(plv["plot_last_histo_ye"] + 1, 1)

    X = plv["plot_area_x0"]

    m_p_e = [ (X - 30) ]
    for chro in plv["plot_chros"]:

        c_l = byc["cytolimits"][str(chro)]
        chr_w = c_l["size"] * plv["plot_b2pf"]

        for m_k, m_v in labs.items():

            c = str(m_v.get("chro", "__na__"))

            if str(chro) != c:
                continue

            s = int(m_v.get("start", 0))
            e = int(m_v.get("end", 0))
            l = m_v.get("label", "")

            m_s = X + s * b2pf
            m_e = X + e * b2pf
            m_w = m_e - m_s
            if m_w < 0.5 and m_w > 0:
                m_w = 0.5
            else:
                m_w = round(m_w, 1)
            m_c = round((m_s + m_e) / 2, 1)
            m_l_l = len(l) * 0.5 * plv["plot_marker_font_size"]
            m_l_s = m_c - 0.5 * m_l_l
            m_l_e = m_c + 0.5 * m_l_l

            found_space = False
            l_i = 0

            for p_e in m_p_e:
                if m_l_s > p_e:
                    found_space = True
                    m_p_e[l_i] = m_l_e
                    break
                l_i += 1

            if found_space is False:
                m_p_e.append(m_l_e)

            if len(m_p_e) > max_lane:
                max_lane = len(m_p_e)

            m_y_e = marker_y_e + l_i * (plv["plot_marker_font_size"] + 1)
            m_h = round(m_y_e - marker_y_0, 1)
            l_y_p = marker_y_e + (l_i + 1) * (plv["plot_marker_font_size"] + 1)

            plv["pls"].append(f'<rect x="{ round(m_s, 1)}" y="{ marker_y_0 }" width="{ round(m_w, 1) }" height="{ m_h }" class="marker" style="opacity: 0.4; " />')
            plv["pls"].append(f'<text x="{ m_c }" y="{ l_y_p }" class="marker">{l}</text>')
            marker_status = True

        X += chr_w
        X += plv["plot_region_gap_width"]

    #--------------------- end chromosome loop --------------------------------#
    
    if max_lane > 0:
        plv["Y"] += max_lane * (plv["plot_marker_font_size"] + 1)
        plv["Y"] += plv["plot_region_gap_width"]
        plv["styles"].append(
            f'.marker {{text-anchor: middle; fill: { plv["plot_marker_font_color"] }; font-size: {plv["plot_marker_font_size"]}px;}}'
        )

    return plv

################################################################################

def _add_labs_from_plot_region_labels(plv, byc):

    r_l_s = plv.get("plot_region_labels", [])
    if len(r_l_s) < 1:
        return plv

    for l in r_l_s:

        l_i = re.split(":", l)
        if len(l_i) < 2:
            continue
        c = l_i.pop(0)
        s_e_i = l_i.pop(0)
        s_e = re.split("-", s_e_i)
        s = s_e.pop(0)
        # TODO: check r'^\d+?$'
        if len(s_e) < 1:
            e = str(int(s)+1)
        else:
            e = s_e.pop(0)
        if len(l_i) > 0:
            label = str(l_i.pop(0))
        else:
            label = ""

        m = _make_marker_object(c, s, e, label)
        plv["plot_labels"].update(m)

    return plv

################################################################################

def _add_labs_from_gene_symbols(plv, byc):

    g_s_s = plv.get("plot_gene_symbols", [])
    if len(g_s_s) < 1:
        return plv

    g_l = []

    for q_g in g_s_s:
        genes, e = retrieve_gene_id_coordinates(q_g, byc)
        g_l += genes

    for f_g in g_l:

        m = _make_marker_object(
            f_g.get("reference_name", False),
            f_g.get("start", False),
            f_g.get("end", False),
            f_g.get("symbol", False)
        )

        if m is not False:
            plv["plot_labels"].update(m)

    return plv

################################################################################

def _make_marker_object(chromosome, start, end, label=""):

    m = False

    # Checks here or upstream?
    if False in [chromosome, start, end, label]:
        return m

    m_k = f'{chromosome}:{start}-{end}:{label}'

    m = {
        m_k: {
            "chro": chromosome,
            "start": start,
            "end": end,
            "label": label
        }
    }

    return m

################################################################################

def _plot_add_footer(plv, results, byc):

    today = datetime.date.today()
    x_a_0 = plv["plot_area_x0"]
    x_c_e = x_a_0 + plv["plot_area_width"]

    plv["styles"].append(
        f'.footer-r {{text-anchor: end; fill: { plv["plot_footer_font_color"] }; font-size: {plv["plot_footer_font_size"]}px;}}'
    )   
    plv["styles"].append(
        f'.footer-l {{text-anchor: start; fill: { plv["plot_footer_font_color"] }; font-size: {plv["plot_footer_font_size"]}px;}}'
    )   

    plv["Y"] += plv["plot_footer_font_size"]
    plv["pls"].append(f'<text x="{x_c_e}" y="{plv["Y"]}" class="footer-r">&#169; CC-BY 2001 - {today.year} progenetix.org</text>')

    if plv["histogram_number"] == 1:
        s_no = results[0].get("sample_count", "?")
        plv["pls"].append(f'<text x="{x_a_0}" y="{plv["Y"]}" class="footer-l">{s_no} samples</text>')

    return plv

################################################################################

def _plot_add_cytoband_svg_gradients(plv):

    plv["pls"].insert(0, 
"""
<linearGradient id="{0}gpos100" x1="0%" x2="0%" y1="0%" y2="80%" spreadMethod="reflect">
    <stop offset="0%" stop-color="rgb(39,39,39)" />
    <stop offset="100%" stop-color="rgb(0,0,0)" />
</linearGradient>
<linearGradient id="{0}gpos75" x1="0%" x2="0%" y1="0%" y2="80%" spreadMethod="reflect">
    <stop offset="0%" stop-color="rgb(87,87,87)" />
    <stop offset="100%" stop-color="rgb(39,39,39)" />
</linearGradient>
<linearGradient id="{0}gpos50" x1="0%" x2="0%" y1="0%" y2="80%" spreadMethod="reflect">
    <stop offset="0%" stop-color="rgb(196,196,196)" />
    <stop offset="100%" stop-color="rgb(111,111,111)" />
</linearGradient>
<linearGradient id="{0}gpos25" x1="0%" x2="0%" y1="0%" y2="80%" spreadMethod="reflect">
    <stop offset="0%" stop-color="rgb(223,223,223)" />
    <stop offset="100%" stop-color="rgb(196,196,196)" />
</linearGradient>
<linearGradient id="{0}gneg" x1="0%" x2="0%" y1="0%" y2="80%" spreadMethod="reflect">
    <stop offset="0%" stop-color="white" />
    <stop offset="100%" stop-color="rgb(223,223,223)" />
</linearGradient>
<linearGradient id="{0}gvar" x1="0%" x2="0%" y1="0%" y2="80%" spreadMethod="reflect">
    <stop offset="0%" stop-color="rgb(196,196,196)" />
    <stop offset="100%" stop-color="rgb(111,111,111)" />
</linearGradient>
<linearGradient id="{0}stalk" x1="0%" x2="0%" y1="0%" y2="80%" spreadMethod="reflect">
    <stop offset="0%" stop-color="rgb(39,39,39)" />
    <stop offset="100%" stop-color="rgb(0,0,0)" />
</linearGradient>
<linearGradient id="{0}acen" x1="0%" x2="0%" y1="0%" y2="80%" spreadMethod="reflect">
    <stop offset="0%" stop-color="rgb(163,55,247)" />
    <stop offset="100%" stop-color="rgb(138,43,226)" />
</linearGradient>
""".format(plv["plot_id"])
    )

    return plv




import re
from pymongo import MongoClient

from cgi_parsing import *

################################################################################

def select_dataset_ids(byc):

    if not "dataset_ids" in byc.keys():
        byc.update( { "dataset_ids": [ ] } )

    if ds_id_from_rest_path_value(byc) is not False:
        return byc

    if ds_id_from_accessid(byc) is not False:
        return byc            

    if ds_ids_from_form(byc) is not False:
        return byc            
    
    if ds_ids_from_args(byc) is not False:
        return byc            

    return byc

################################################################################

def ds_id_from_rest_path_value(byc):

    ds_id = rest_path_value("datasets")
    if ds_id is "empty_value":
        return False

    if ds_id not in byc["dataset_definitions"].keys():
        return False

    byc.update( { "dataset_ids": [ ds_id ] } )
    return byc

################################################################################

def ds_id_from_accessid(byc):

    # TODO: This is very verbose. In principle there should be an earlier
    # test of existence...

    accessid = byc["form_data"].get("accessid", None)
    if "accessid" is None:
        return False

    info_db = byc["config"].get("info_db", None)
    if "info_db" is None:
        return False

    ho_collname = byc["config"].get("handover_coll", None)
    if "ho_collname" is None:
        return False

    ho_client = MongoClient()
    h_o = ho_client[ info_db ][ ho_collname ].find_one( { "id": accessid } )
    if not h_o:
        return False

    ds_id = h_o.get("source_db", None)
    if "ds_id" is None:
        return False

    if ds_id not in byc["dataset_definitions"].keys():
        return False

    byc.update( { "dataset_ids": [ ds_id ] } )
    return byc

################################################################################

def ds_ids_from_form(byc):

    f_ds_ids = byc["form_data"].get("dataset_ids", None)
    if f_ds_ids is None:
        return False

    ds_ids = [ ]
    for ds_id in f_ds_ids:
        if ds_id in byc["dataset_definitions"].keys():
            ds_ids.append(ds_id)

    if len(ds_ids) < 1:
        return False

    byc.update( { "dataset_ids": ds_ids } )
    
    return byc

################################################################################

def ds_ids_from_args(byc):

    if not "args" in byc or byc["args"] is None:
        return byc

    if byc["args"].datasetIds:
        ds_ids = re.split(",", byc["args"].datasetIds)
        byc.update( { "dataset_ids": ds_ids } )
        return byc

    return byc



from os import environ, path, pardir
import inspect, json, random
from pathlib import Path
from pymongo import MongoClient
from bson import json_util

# local
bycon_lib_path = path.dirname( path.abspath(__file__) )
pkg_path = path.join( bycon_lib_path, pardir )

from args_parsing import *
from cgi_parsing import *
from cytoband_utils import translate_reference_ids
from data_retrieval import *
from dataset_parsing import select_dataset_ids
from datatable_utils import export_datatable_download
from export_file_generation import *
from handover_generation import dataset_response_add_handovers, query_results_save_handovers, dataset_results_save_handovers
from interval_utils import generate_genomic_intervals, interval_counts_from_callsets
from plot_utils import histoplot_svg_generator
from query_execution import execute_bycon_queries, process_empty_request, mongo_result_list
from query_generation import  generate_queries, initialize_beacon_queries, paginate_list, set_pagination_range
from read_specs import load_yaml_empty_fallback, read_bycon_configs_by_name, read_bycon_definition_files, read_local_prefs
from response_remapping import *
from schema_parsing import *

################################################################################

def initialize_bycon(config):

    b_r_p = config.get("byc_root_pars", {})

    byc =  {
        "pkg_path": pkg_path,
        "bycon_lib_path": bycon_lib_path
    }

    for k, v in b_r_p.items():
        byc.update({k:v})

    config.pop("byc_root_pars", None)
    byc.update({"config":config})

    if not environ.get('HTTP_HOST'):
        byc.update({"env": "local"})

    return byc

################################################################################

def beacon_data_pipeline(byc, entry_type):

    initialize_bycon_service(byc, entry_type)

    run_beacon_init_stack(byc)
    return_filtering_terms_response(byc)
    run_result_sets_beacon(byc)
    update_meta_queries(byc)
    query_results_save_handovers(byc)
    check_computed_histoplot_delivery(byc)
    check_computed_interval_frequency_delivery(byc)
    check_switch_to_count_response(byc)
    check_switch_to_boolean_response(byc)
    cgi_print_response( byc, 200 )

################################################################################

def run_beacon_init_stack(byc):

    select_dataset_ids(byc)
    if len(byc["dataset_ids"]) < 1:
        print_text_response("No existing dataset_id - please check dataset_definitions")

    create_empty_beacon_response(byc)
    initialize_beacon_queries(byc)
    generate_genomic_intervals(byc)
    response_collect_errors(byc)
    cgi_break_on_errors(byc)

    return byc

################################################################################

def initialize_bycon_service(byc, service=False):

    """For consistency, the name of the local configuration file should usually
    correspond to the calling main function. However, an overwrite can be
    provided."""

    defs = byc.get("beacon_defaults", {})
    b_e_d = defs.get("entity_defaults", {})    
    form = byc["form_data"]

    frm = inspect.stack()[1]
    if service is False:
        service = frm.function

    # TODO - streamline
    s_a_s = byc["beacon_mappings"].get("service_aliases", {})

    if service in s_a_s:
        service = s_a_s[service]

    service = decamelize(service)

    mod = inspect.getmodule(frm[0])
    if mod is not None:
        """
        Here we allow the addition of additional configuration files, necessary
        for options beyond basic library use. Files are read in as
        1. from a `config` directory in the parent directory of the executed script
            * this e.g. overwrites the config options
        2. from a config directory inside the script directory
            * this is usually used to provide script-specific parameters (`service_defaults`...)
        """
        sub_path = path.dirname( path.abspath(mod.__file__) )

        conf_dir = path.join( sub_path, "local" )
        if path.isdir(conf_dir):
            c_f = Path( path.join( conf_dir, "config.yaml" ) )
            if path.isfile(c_f):
                byc.update({"config": load_yaml_empty_fallback( c_f ) })
            d_f = Path( path.join( conf_dir, "beacon_defaults.yaml" ) )
            if path.isfile(d_f):
                byc.update({"beacon_defaults": load_yaml_empty_fallback( d_f ) })
                defaults = byc["beacon_defaults"].get("defaults", {})
                for d_k, d_v in defaults.items():
                    byc.update( { d_k: d_v } )
            read_bycon_definition_files(conf_dir, byc)

        read_local_prefs( service, sub_path, byc )

    get_bycon_args(byc)
    args_update_form(byc)

    conf = byc["service_config"]

    if "defaults" in conf:
        for d_k, d_v in conf["defaults"].items():
            byc.update( { d_k: d_v } )

    if service in b_e_d:
        for d_k, d_v in b_e_d[ service ].items():
            byc.update( { d_k: d_v } )

    # update response_entity_id from path
    update_entity_ids_from_path(byc)

    # update response_entity_id from form
    update_requested_schema_from_request(byc)
    set_response_entity(byc)

    set_io_params(byc)
    translate_reference_ids(byc)
    set_special_modes(byc)

    # TODO: standardize the general defaults / entity defaults / form values merging
    #       through pre-parsing into identical structures and then use deepmerge etc.

    return byc

################################################################################

def set_special_modes(byc):

    form = byc["form_data"]
    for m in ["test_mode", "debug_mode", "download_mode", "include_handovers"]:
        if m in form:
            byc.update({m: test_truthy( form.get(m, False) ) })

    t_m_k = form.get("test_mode_count", "___none___")
    if re.match(r'^\d+$', str(t_m_k)):
        byc.update({"test_mode_count": int(t_m_k) })

    return byc

################################################################################

def set_io_params(byc):

    form = byc["form_data"]

    if not "pagination" in byc:
        byc.update( { "pagination": {"skip": 0, "limit": 0 } } )

    for sp in ["skip", "limit"]:
        if sp in form:
            if re.match(r'^\d+$', str(form[sp])):
                s_v = int(form[sp])
                byc["pagination"].update({sp: s_v})

    byc.update({"output": form.get("output", byc["output"])})

    # TODO: this is only used in some services ...
    if "method_keys" in byc["service_config"]:
        m = form.get("method", "___none___")
        if m in byc["service_config"]["method_keys"].keys():
            byc["method"] = m

    return byc

################################################################################

def update_entity_ids_from_path(byc):

    if not byc["request_entity_path_id"]:
        return byc

    if not byc["response_entity_path_id"]:
        byc.update({ "response_entity_path_id": byc["request_entity_path_id"] })

    req_p_id = byc["request_entity_path_id"]
    res_p_id = byc["response_entity_path_id"]

    m_k = byc["request_path_root"] + "_mappings"
    b_mps = byc[m_k]
    p_r_m = b_mps["path_response_type_mappings"]

    byc.update({
        "request_entity_id": p_r_m.get(req_p_id, byc["request_entity_path_id"]),
        "response_entity_id": p_r_m.get(res_p_id, byc["response_entity_path_id"])
    })

    return byc

################################################################################

def update_requested_schema_from_request(byc):

    m_k = byc["request_path_root"] + "_mappings"
    b_mps = byc[m_k]
    form = byc["form_data"]
    b_qm = byc["query_meta"]

    if "requested_schema" in form:
        byc.update({"response_entity_id": form.get("requested_schema", byc["response_entity_id"])})
    elif "requested_schemas" in b_qm:
        byc.update({"response_entity_id": b_qm["requested_schemas"][0].get("entity_type", byc["response_entity_id"])})

    return byc

################################################################################

def set_response_entity(byc):

    m_k = byc["request_path_root"] + "_mappings"
    b_rt_s = byc[m_k]["response_types"]

    if byc["response_entity_id"] in b_rt_s.keys():
        byc.update({"response_entity": b_rt_s[ byc["response_entity_id"] ] })

    return byc

################################################################################

def run_result_sets_beacon(byc):

    sr_r = byc["service_response"]["response"]

    byc.update({"dataset_results":{}})

    ######## result sets loop ##################################################

    for i, r_set in enumerate(sr_r["result_sets"]):

        ds_id = r_set["id"]
        generate_queries(byc, ds_id)

        r_set, r_s_res = populate_result_set(r_set, byc)

        if r_set["results_count"] < 1:
            continue

        # TODO: This condition avoids the "range on range" if previously set for
        # a handover query (by accessioinId)
        if not "range" in byc["pagination"]:
            set_pagination_range(len(r_s_res), byc)        
        if test_truthy( byc["form_data"].get("paginate_results", True) ):
            r_s_res = paginate_list(r_s_res, byc)

        check_alternative_single_set_deliveries(ds_id, r_s_res, byc)
        r_s_res = reshape_resultset_results(ds_id, r_s_res, byc)

        sr_r["result_sets"][i].update({
            "paginated_results_count": len( r_s_res ),
            "exists": True,
            "results": r_s_res
        })

    ######## end of result sets loop ###########################################

    sr_rs = byc["service_response"].get("response_summary", {})
    sr_rs.update({"num_total_results":0})
    for r_set in sr_r["result_sets"]:
        sr_rs["num_total_results"] += r_set["results_count"]

    if sr_rs["num_total_results"] > 0:
        sr_rs.update({"exists": True })

    return byc

################################################################################

def populate_result_set(r_set, byc):

    ds_id = r_set["id"]
    execute_bycon_queries( ds_id, byc )

    # # Special check-out here since this forces a single handover
    # check_plot_responses(ds_id, byc)

    r_set.update({ "results_handovers": dataset_response_add_handovers(ds_id, byc) })
    r_s_res = retrieve_data(ds_id, byc)

    r_set_update_counts(r_set, r_s_res, byc)

    return r_set, r_s_res

################################################################################

def r_set_update_counts(r_set, r_s_res, byc):

    ds_id = r_set["id"]
    ds_res = byc["dataset_results"][ds_id]

    r_set.update({"results_count": 0})

    for c, c_d in byc["config"]["beacon_counts"].items():
        h_o_k = c_d.get("h->o_key", None)
        if h_o_k is None:
            continue
        if h_o_k not in ds_res:
            continue
        r_c = ds_res[ h_o_k ]["target_count"]
        s_c = len(list(set(ds_res[ h_o_k ]["target_values"])))
        r_set["info"]["counts"].update({ c: r_c })

    if byc["empty_query_all_count"]:
        r_set.update({"results_count": byc["empty_query_all_count"] })
    elif isinstance(r_s_res, list):
        r_set.update({"results_count": len( r_s_res ) })

    return r_set

################################################################################

def check_alternative_single_set_deliveries(ds_id, r_s_res, byc):

    check_datatable_delivery(r_s_res, byc)
    check_alternative_variant_deliveries(ds_id, byc)
    check_alternative_callset_deliveries(ds_id, byc)

    return byc

################################################################################

def response_meta_add_request_summary(r, byc):

    if not "received_request_summary" in r["meta"]:
        return r

    r_rcvd_rs = r["meta"]["received_request_summary"]
    defs = byc.get("beacon_defaults", {})
    b_e_d = defs.get("entity_defaults", {})

    form = byc["form_data"]

    r_rcvd_rs.update({
        "filters": byc.get("filters", []), 
        "pagination": byc.get("pagination", {}),
        "api_version": b_e_d["info"].get("api_version", "v2")
    })

    for p in ["include_resultset_responses", "requested_granularity"]:
        if p in form:
            r_rcvd_rs.update({p:form.get(p)})
            if "requested_granularity" in p:
                r["meta"].update({"returned_granularity": form.get(p)})

    try:
        for rrs_k, rrs_v in byc["service_config"]["meta"]["received_request_summary"].items():
            r_rcvd_rs.update( {rrs_k: rrs_v })
    except:
        pass

    return r

################################################################################

def update_meta_queries(byc):

    try:
        if not "info" in byc["meta"]:
            byc["meta"].update({"info":{}})
        byc["service_response"]["meta"]["info"].update({ "processed_query": byc.get("original_queries", {}) })
    except:
        pass

    return byc

################################################################################

def create_empty_beacon_response(byc):

    r_s = byc["response_entity"].get("response_schema", None)
    if r_s is None:
        return byc

    r, e = instantiate_response_and_error(byc, r_s)
    response_update_meta(r, byc)

    try:
        r["response"].update({"result_sets":[]})
    except KeyError:
        pass

    if "beaconResultsetsResponse" in r_s:

        # TODO: stringent definition on when this is being used
        r_set = object_instance_from_schema_name(byc, "beaconResultsets", "definitions/ResultsetInstance/properties")

        if "dataset_ids" in byc:
            for ds_id in byc[ "dataset_ids" ]:
                ds_rset = r_set.copy()
                ds_rset.update({
                    "id": ds_id,
                    "set_type": "dataset",
                    "results_count": 0,
                    "exists": False,
                    "info": { "counts": { } }
                })
                r["response"]["result_sets"].append(ds_rset)

    byc.update( {"service_response": r, "error_response": e })

    # saving the parameters to the response
    for p in ["method", "dataset_ids", "filters", "variant_pars"]:
        if p in byc:
            response_add_received_request_summary_parameter(byc, p, byc[ p ])

    return byc

################################################################################

def create_empty_service_response(byc):

    r, e = instantiate_response_and_error(byc, byc["response_entity"]["response_schema"])
    response_update_meta(r, byc)

    byc.update( {"service_response": r, "error_response": e })

    # saving the parameters to the response
    for p in ["method", "dataset_ids", "filters", "variant_pars"]:
        if p in byc:
            response_add_received_request_summary_parameter(byc, p, byc[ p ])

    return byc

###############################################################################

def create_empty_non_data_response(byc):

    r, e = instantiate_response_and_error(byc, byc["response_entity"]["response_schema"])
    response_update_meta(r, byc)

    for r_k, r_v in byc["service_config"]["response"].items():
        r["response"].update({r_k: r_v})

    byc.update( {"service_response": r, "error_response": e })

    return byc

################################################################################

def check_plot_responses(ds_id, byc):

    if not "plot" in byc["output"]:
        return byc

    check_cnvhistogram_plot_response(ds_id, byc)

    return byc

################################################################################

def check_cnvhistogram_plot_response(ds_id, byc):

    h_o_types = byc["handover_definitions"]["h->o_types"]
    ds_h_o = byc["dataset_definitions"][ ds_id ].get("handoverTypes", h_o_types.keys())

    if "cnvhistogram" not in ds_h_o:
        return

    byc["include_handovers"] = True
    byc["dataset_definitions"][ ds_id ].update({"handoverTypes": ["cnvhistogram"] })
    dataset_results_save_handovers(ds_id, byc)
    r_s_ho = dataset_response_add_handovers(ds_id, byc)

    for h_o in r_s_ho:
        if "cnvhistogram" in h_o["handover_type"]["id"]:
            print_uri_rewrite_response(h_o["url"], "", "")
            exit()

    return byc

################################################################################

def check_datatable_delivery(results, byc):

    if not "table" in byc["output"]:
        return byc
    if not "datatable_mappings" in byc:
        return byc

    export_datatable_download(results, byc)

################################################################################

def instantiate_response_and_error(byc, schema):

    """The response relies on the pre-processing of input parameters (queries etc)."""
    r = object_instance_from_schema_name(byc, schema, "properties")
    m = object_instance_from_schema_name(byc, "beaconResponseMeta", "properties")
    r.update({"meta": m })
    if byc["debug_mode"] is True:
        print(byc["response_entity"]["response_schema"])
        prjsonnice(r)
    e = object_instance_from_schema_name(byc, "beaconErrorResponse", "properties")

    return r, e

################################################################################

def response_update_meta(r, byc):

    if "response_summary" in r:
        r["response_summary"].update({ "exists": False })
    response_meta_set_info_defaults(r, byc)
    response_meta_set_config_defaults(r, byc)
    response_meta_set_entity_values(r, byc)
    response_meta_add_request_summary(r, byc)
    r["meta"].update({"test_mode": byc.get("test_mode", False)})

################################################################################

def response_meta_set_info_defaults(r, byc):

    defs = byc.get("beacon_defaults", {})
    b_e_d = defs.get("entity_defaults", {})

    for i_k in ["api_version", "beacon_id"]:
        r["meta"].update({ i_k: b_e_d["info"].get(i_k, "") })

    return r

################################################################################

def response_meta_set_config_defaults(r, byc):

    # TODO: should be in schemas
    m = byc["service_config"].get("meta", {})
    for k, v in m.items():
        r["meta"].update( { k: v } )

    return r

################################################################################

def response_meta_set_entity_values(r, byc):
 
    try:
        r["meta"]["received_request_summary"].update({
            "requested_schemas": [ byc["response_entity"]["beacon_schema"] ]
        })
        r["meta"].update( { "returned_schemas": [ byc["response_entity"]["beacon_schema"] ] } )
    except:
        pass

    return r

################################################################################

def response_add_received_request_summary_parameter(byc, name, value):

    # TODO: proper request parameters
    if "variant_pars" in name:
        name = "request_parameters"

    if not "received_request_summary" in byc["service_response"]["meta"]:
        return byc

    if value:
        byc["service_response"]["meta"]["received_request_summary"].update( { name: value } )

    return byc

################################################################################

def response_collect_errors(byc):

    # TODO: flexible list of errors
    if not byc[ "queries" ].keys():
      response_add_error(byc, 200, "No (correct) query parameters were provided." )
    if len(byc[ "dataset_ids" ]) > 1:
      response_add_error(byc, 200, "More than 1 `datasetIds` value was provided." )
      
################################################################################

def response_add_error(byc, code=200, message=False):

    if message is False:
        return byc
    if len(str(message)) < 1:
        return byc

    e = { "error_code": code, "error_message": message }
    byc["error_response"].update({"error": e})

    return byc

################################################################################

def response_add_warnings(byc, message=False):

    if message is False:
        return byc
    if len(str(message)) < 1:
        return byc

    if not "service_response" in byc:
        return byc

    if not "info" in byc["service_response"]:
        byc["service_response"].update({"info": {}})
    if not "warnings" in byc["service_response"]:
        byc["service_response"]["info"].update({"warnings": []})

    byc["service_response"]["info"]["warnings"].append(message)

    return byc

################################################################################

def collations_set_delivery_keys(byc):

    # the method keys can be overriden with "deliveryKeys"

    method = byc.get("method", False)
    conf = byc["service_config"]

    d_k = form_return_listvalue( byc["form_data"], "deliveryKeys" )

    if len(d_k) > 0:
        return d_k

    if method is False:
        return d_k

    # here the empty list is returned, leading to all
    if "details" in method:
        return d_k

    if not "method_keys" in conf:
        return d_k

    if method in conf["method_keys"]:
        d_k = conf["method_keys"][ method ]

    return d_k

################################################################################

def populate_service_response( byc, results):

    byc["service_response"].update({
        "response_summary": {
            "exists": False,
            "num_total_results": 0
        }
    })

    populate_service_header(byc, results)
    populate_service_response_counts(byc)

    if len(results) > 0:
        byc["service_response"]["response_summary"].update({
            "exists": True,
            "num_total_results": len(results)
        })

    byc["service_response"]["response"].update({"results": results })

    check_switch_to_count_response(byc)
    check_switch_to_boolean_response(byc)

    return byc

################################################################################

def populate_service_header(byc, results):

    try:
        r_s = byc["service_response"]["response_summary"]
    except:
        return byc
    r_no = 0

    if isinstance(results, list):
        r_no = len( results )
        r_s.update({"num_total_results": r_no })
    if r_no > 0:
        r_s.update({"exists": True })

    return byc

################################################################################

def populate_service_response_counts(byc):

    if not "dataset_results" in byc:
        return byc
    if not "dataset_ids" in byc:
        return byc

    ds_id = byc["dataset_ids"][0]

    if not ds_id in byc["dataset_results"]:
        return byc

    counts = { }

    for c, c_d in byc["config"]["beacon_counts"].items():

        if c_d["h->o_key"] in byc["dataset_results"][ds_id]:
            counts[ c ] = byc["dataset_results"][ds_id][ c_d["h->o_key"] ]["target_count"]

    byc["service_response"]["info"].update({ "counts": counts })

    return byc

################################################################################

def check_alternative_variant_deliveries(ds_id, byc):

    if not "genomicVariant" in byc["response_entity_id"]:
        return byc

    if "pgxseg" in byc["output"]:
        export_pgxseg_download(ds_id, byc)

    if "variants" in byc["method"]:
        export_variants_download(ds_id, byc)

    return byc

################################################################################

def check_alternative_callset_deliveries(ds_id, byc):

    check_callsets_matrix_delivery(ds_id, byc)

    return byc

################################################################################

def return_filtering_terms_response( byc ):

    if not "filteringTerm" in byc["response_entity_id"]:
        return byc

    # TODO: correct response w/o need to fix
    byc["service_response"].update({"response": { "filteringTerms": [], "resources": []} })

    f_r_d = {}

    f_db = byc["config"]["info_db"]
    f_coll = byc["config"]["collations_coll"]

    f_t_s = [ ]
    ft_fs = [ ]

    if "filters" in byc:
        if len(byc["filters"]) > 0:
            for f in byc["filters"]:
                ft_fs.append('('+f["id"]+')')
    f_s = '|'.join(ft_fs)
    f_re = re.compile(r'^'+f_s)

    collation_types = set()

    for ds_id in byc[ "dataset_ids" ]:

        query = { "dataset_id": ds_id }

        try:
            if len(byc["form_data"]["scope"]) > 4:
                query.update({"scope": byc["form_data"]["scope"]})
        except:
            pass

        fields = { "_id": 0 }

        f_s, e = mongo_result_list(f_db, f_coll, query, fields)

        t_f_t_s = [ ]

        for f in f_s:
            collation_types.add(f.get("collation_type", None))
            f_t = { "count": f.get("count", 0)}
            for k in ["id", "type", "label"]:
                if k in f:
                    f_t.update({k:f[k]})
            t_f_t_s.append(f_t)

        f_t_s.extend(t_f_t_s)

    byc["service_response"]["response"].update({ "filteringTerms": f_t_s })
    byc["service_response"]["response"].update({ "resources": create_filters_resource_response(collation_types, byc) })

    cgi_print_response( byc, 200 )

################################################################################

def create_filters_resource_response(collation_types, byc):

    r_o = {}
    resources = []

    f_d_s = byc["filter_definitions"]
    collation_types = list(collation_types)
    res_schema = object_instance_from_schema_name(byc, "beaconFilteringTermsResults", "definitions/Resource/properties", "json")    
    for c_t in collation_types:
        f_d = f_d_s[c_t]
        r = {}
        for k in res_schema.keys():
            if k in f_d:
                r.update({k:f_d[k]})

        r_o.update( {f_d["namespace_prefix"]: r })

    for k, v in r_o.items():
        resources.append(v)
    return resources

################################################################################

def check_computed_histoplot_delivery(byc):

    if byc["output"] not in ["histoplot", "cnvhistogram"]:
        return byc

    p_r = byc["pagination"]
    interval_sets = []

    for ds_id in byc["dataset_results"].keys():

        ds_results = byc["dataset_results"][ds_id]

        if not "callsets._id" in ds_results:
            continue

        cs_r = ds_results["callsets._id"]

        mongo_client = MongoClient()
        cs_coll = mongo_client[ ds_id ][ "callsets" ]
        q_vals = cs_r["target_values"]
        r_no = len(q_vals)
        if r_no > p_r["limit"]:
            q_vals = paginate_list(q_vals, byc)

        cs_cursor = cs_coll.find({"_id": {"$in": q_vals }, "variant_class": { "$ne": "SNV" } } )
        intervals, cnv_cs_count = interval_counts_from_callsets(cs_cursor, byc)
        # print(f"{intervals[88]}")

        i_set = {
            "dataset_id": ds_id,
            "group_id": ds_id,
            "label": "Search Results",
            "sample_count": cnv_cs_count,
            "interval_frequencies": []
        }

        for intv_i, intv in enumerate(intervals):
            i_set["interval_frequencies"].append(intv.copy())

        interval_sets.append(i_set)

    histoplot_svg_generator(byc, interval_sets)

################################################################################

def check_computed_interval_frequency_delivery(byc):

    if not "frequencies" in byc["output"]:
        return byc

    ds_id = byc[ "dataset_ids" ][ 0 ]
    ds_results = byc["dataset_results"][ds_id]
    p_r = byc["pagination"]

    if not "callsets._id" in ds_results:
        return byc

    cs_r = ds_results["callsets._id"]

    mongo_client = MongoClient()
    cs_coll = mongo_client[ ds_id ][ "callsets" ]

    open_text_streaming(byc["env"], "interval_cnv_frequencies.pgxseg")

    for d in ["id", "assemblyId"]:
        print("#meta=>{}={}".format(d, byc["dataset_definitions"][ds_id][d]))
    print_filters_meta_line(byc)
    print('#meta=>query="{}"'.format(json.dumps(byc["queries_at_execution"], indent=None, sort_keys=True, default=str)))
    print('#meta=>skip={};limit={}'.format(p_r["skip"], p_r["limit"]))
    print("#meta=>sample_no={}".format(cs_r["target_count"]))

    q_vals = cs_r["target_values"]
    r_no = len(q_vals)
    if r_no > p_r["limit"]:
        q_vals = paginate_list(q_vals, byc)
        print('#meta=>"WARNING: Only analyses {} - {} (out of {}) used for calculations due to pagination skip and limit"'.format((p_r["range"][0] + 1), p_r["range"][-1], cs_r["target_count"]))

    h_ks = ["reference_name", "start", "end", "gain_frequency", "loss_frequency", "no"]
    print("group_id\t"+"\t".join(h_ks))

    cs_cursor = cs_coll.find({"_id": {"$in": q_vals }, "variant_class": { "$ne": "SNV" } } )

    intervals, cnv_cs_count = interval_counts_from_callsets(cs_cursor, byc)

    for intv in intervals:
        v_line = [ ]
        v_line.append("query_result")
        for k in h_ks:
            v_line.append(str(intv[k]))
        print("\t".join(v_line))

    close_text_streaming()

################################################################################

def check_callsets_matrix_delivery(ds_id, byc):

    if not "pgxmatrix" in byc["output"]:
        return byc

    export_callsets_matrix(ds_id, byc)
        